

#include "R3000A_Recompiler.h"

using namespace R3000A;


#define _USE_LOAD_CALLBACK
#define _USE_BRANCH_CALLBACK

x64Encoder *Recompiler::e;
//ICache_Device *Recompiler::ICache;
Cpu *Recompiler::r;
s32 Recompiler::OpLevel;
u32 Recompiler::LocalPC;
u32 Recompiler::Local_LastModifiedReg;
u32 Recompiler::Local_NextPCModified;

u32 Recompiler::CurrentCount;

u32 Recompiler::isBranchDelaySlot;
u32 Recompiler::isLoadDelaySlot;

u32 Recompiler::bStopEncodingAfter;
u32 Recompiler::bStopEncodingBefore;


// constructor
// NumberOfBlocks MUST be a power of 2, so 1 would mean 2, 2 would mean 4
Recompiler::Recompiler ( Cpu* R3000ACpu, u32 NumberOfBlocks, u32 BlockSize_PowerOfTwo, u32 MaxIStep_Shift )
{
	Reset ();
	
	
	BlockSize = 1 << BlockSize_PowerOfTwo;
	
	MaxStep_Shift = MaxIStep_Shift;
	MaxStep = 1 << MaxIStep_Shift;
	MaxStep_Mask = MaxStep - 1;
	
	NumBlocks = 1 << NumberOfBlocks;
	NumBlocks_Mask = NumBlocks - 1;
	
	// allocate variables
	StartAddress = new u32 [ NumBlocks ];
	//LastAddress = new u32 [ NumBlocks ];
	RunCount = new u32 [ NumBlocks ];
	MaxCycles = new u64 [ NumBlocks ];
	Instructions = new u32 [ NumBlocks * MaxStep ];
	
	// initialize the address and instruction so it is known that it does not refer to anything
	memset ( StartAddress, 0xff, sizeof( u32 ) * NumBlocks );
	//memset ( LastAddress, 0xff, sizeof( u32 ) * NumBlocks );
	memset ( RunCount, 0xff, sizeof( u32 ) * NumBlocks );
	memset ( Instructions, 0xff, sizeof( u32 ) * NumBlocks * MaxStep );
	
	// create the encoder
	//e = new x64Encoder ( BlockSize_PowerOfTwo, NumBlocks );
	InstanceEncoder = new x64Encoder ( BlockSize_PowerOfTwo, NumBlocks );
	
	//ICache = IC;
	r = R3000ACpu;
}



// destructor
Recompiler::~Recompiler ()
{
	delete e;
	
	// delete the variables that were allocated
	delete StartAddress;
	//delete LastAddress;
	delete RunCount;
	delete MaxCycles;
	delete Instructions;
}


void Recompiler::Reset ()
{
	memset ( this, 0, sizeof( Recompiler ) );	
}


u32 Recompiler::CloseOpLevel ( u32 OptLevel, u32 Address )
{
	switch  ( OptLevel )
	{
		case 0:
			break;
			
		case 1:
			// write back last modified register if in load delay slot
			//if ( isLoadDelaySlot )
			//{
				e->MovMemImm32 ( &r->LastModifiedRegister, Local_LastModifiedReg );
			//}
			
			// write back "NextPC" if there was no SYSCALL
			if ( !Local_NextPCModified )
			{
				e->MovMemImm32 ( &r->NextPC, Address );
			}
			break;
			
		case 2:
			// write back last modified register if in load delay slot
			//if ( isLoadDelaySlot )
			//{
				e->MovMemImm32 ( &r->LastModifiedRegister, Local_LastModifiedReg );
			//}
			
			// write back "NextPC" if there was no SYSCALL
			if ( !Local_NextPCModified )
			{
				e->MovMemImm32 ( &r->NextPC, Address );
			}
			break;
	}
}

// returns number of instructions that were recompiled
u32 Recompiler::Recompile ( u32 BeginAddress, void* SrcCode )
{
	u32 Address, Block;
	s32 ret, Cycles;
	Instruction::Format inst;
	
	u32 RecompileCount;
	s32 MaxCount;
	
	u32* pInstrPtr;
	u32* pSrcCodePtr;
	
	// mask address
	// don't do this
	//StartAddress &= c_iAddress_Mask;
	
	// set the encoder to use
	e = InstanceEncoder;
	
	// save the address?
	Address = BeginAddress;
	
	// set the current optimization level
	OpLevel = OptimizeLevel;
	
	// get the block to encode in
	Block = ( BeginAddress >> 2 ) & NumBlocks_Mask;
	
	// set block initially to cache
	//DoNotCache [ Block ] = 0;
	
	// start in code block
	e->StartCodeBlock ( Block );
	
	// set the start address for code block
	// address must actually match exactly. No mask
	//StartAddress [ Block ] = BeginAddress & c_iAddress_Mask;
	StartAddress [ Block ] = BeginAddress;
	
	// set the instruction
	//Instructions [ Block ] = *((u32*) SrcCode);
	pInstrPtr = & ( Instructions [ Block << MaxStep_Shift ] );
	pSrcCodePtr = (u32*) SrcCode;
	
	// clear cache miss count
	//CacheMissCount [ Block ] = 0;
	
	// start cycles at zero
	Cycles = 0;
	
	// start PC
	//LocalPC = r->PC;
	
	// init count of recompiled instructions
	RecompileCount = 0;
	
	// NextPC has not been modified yet
	Local_NextPCModified = false;
	
	// want to stop at cache boundaries (would need extra code there anyways)
	MaxCount = MaxStep - ( ( Address >> 2 ) & MaxStep_Mask );
	if ( MaxCount <= 0 ) MaxCount = 1;
	
	// some instructions need to stop encoding either before or after the instruction, at least for now
	// if stopping before, it keeps the instruction if there is nothing before it in the run
	bStopEncodingAfter = false;
	bStopEncodingBefore = false;
	
	//for ( int i = 0; i < MaxStep; i++, Address += 4 )
	for ( int i = 0; i < MaxCount; i++ )
	{
			// start encoding a MIPS instruction
			e->StartInstructionBlock ();
			
			// set r0 to zero for now
			//e->MovMemImm32 ( &r->GPR [ 0 ].u, 0 );
			
			// in front of the instruction, set NextPC to the next instruction
			// do this at beginning of code block
			// NextPC = PC + 4
			//e->MovMemImm32 ( &r->NextPC, Address + 4 );
			
			
			// recompile the instruction
			//inst.Value = *((u32*) SrcCode);
			inst.Value = *pSrcCodePtr;
			ret = Recompile ( inst, Address );
			
			
			if ( ret <= 0 )
			{
				// there was a problem, and recompiling is done
				
				// need to undo whatever we did for this instruction
				e->UndoInstructionBlock ();
				
	//cout << "\nUndo: Address=" << hex << Address;
				
				// TODO: if no instructions have been encoded yet, then just try again with a lower optimization level
				if ( ( OpLevel > 0 ) && ( !RecompileCount ) )
				{
	//cout << "\nNext Op Level down";			
					OpLevel--;
					//Address -= 4;
					i--;
					
					// at this point, this should be the last instruction since we had to go down an op level
					MaxCount = 1;
				}
				else
				{
				
					//cout << "\nhps1x64: R3000A: Recompiler: Error: Unable to encode instruction.";
					
					// mark block as unable to recompile if there were no instructions recompiled at all
					//if ( !Cycles ) DoNotCache [ Block ] = 1;
					
					// done
					break;
				}
			}
			else
			{
				// instruction successfully encoded from MIPS into x64
				e->EndInstructionBlock ();
				
	//cout << "\nCool: Address=" << hex << Address << " ret=" << dec << ret << " inst=" << hex << *pSrcCodePtr << " i=" << dec << i;
	
				// update number of instructions that have been recompiled
				RecompileCount++;
				
				
				// update to next instruction
				//SrcCode += 4;
				*pInstrPtr++ = *pSrcCodePtr++;
				
				
				// add number of cycles encoded
				Cycles += ret;
				
				// update address
				Address += 4;
				
				// update program counter for recompiling
				//LocalPC += 4;
				
			}
			
		
	} // end for ( int i = 0; i < MaxStep; i++, Address += 4 )
	
	
	// store the last modified register back if optimization level is greater than zero
	if ( OpLevel )
	{
		// write back last modified register if in load delay slot
		//if ( isLoadDelaySlot )
		//{
			e->MovMemImm32 ( &r->LastModifiedRegister, Local_LastModifiedReg );
		//}
	}
	
	// in front of the instruction, set NextPC to the next instruction
	// note: SYSCALL, etc modifies NextPC
	// do this at beginning of code block
	// NextPC = PC + 4
	if ( !Local_NextPCModified )
	{
		e->MovMemImm32 ( &r->NextPC, Address );
	}
	
	// *** todo *** put in the end of code block (set next PC, etc) and return
	// *** todo *** will need to make sure there is at least 8 bytes remaining for the footer (actually 7 bytes, but should check for 8)
	//e->MovMemImm32 ( &r->NextPC, Address );
	e->x64EncodeReturn ();
	
	// set the end address for code block (inclusive)
	// but with no mask, use actual address
	//LastAddress [ Block ] = ( Address - 4 ) & c_iAddress_Mask;
	//LastAddress [ Block ] = ( Address - 4 );
	RunCount [ Block ] = RecompileCount;
	
	// return the number of instructions that were encoded minus 1
	e->MovReg64Imm32 ( RAX, RecompileCount - 1 );
	
	// put in max cycles for block
	// code block is not valid if this is zero
	MaxCycles [ Block ] = Cycles;
	
	// done encoding block
	e->EndCodeBlock ();
	
	//return ret;
	return RecompileCount;
}

//void Recompiler::Invalidate ( u32 Address )
void Recompiler::Invalidate ( u32 Address, u32 Count )
{
	s32 StartBlock, StopBlock;
	
	u32 iBlock;
	
	/*
	// mask addresses??
	Address &= c_iAddress_Mask;
	
	// Get block to stop checking address at. 4 bytes per instruction
	StopBlock = ( Address >> 2 ) & NumBlocks_Mask;
	
	// get block to start checking address at
	StartBlock = StopBlock - ( MaxStep - 1 );
	
	if ( StartBlock < 0 ) StartBlock = 0;
	*/
	
	// get the address to actually start checking from
	Address -= ( MaxStep - 1 );
	
	// update the count
	Count += ( MaxStep - 1 );
	
	// get the starting block
	iBlock = ( Address >> 2 ) & NumBlocks_Mask;
	
	while ( Count > 0 )
	{
		if ( ( StartAddress [ iBlock ] == Address ) && ( RunCount [ iBlock ] >= Count ) )
		{
			// invalidate the block
			StartAddress [ iBlock ] = 0xffffffff;
		}
		
		// update vars
		Address += 4;
		iBlock++;
		Count--;
	}
	
	/*
	for ( u32 i = StartBlock; i <= StopBlock; i++ )
	{
		if ( Address >= StartAddress [ i ] && Address <= LastAddress [ i ] )
		{
			// invalidate any related code cache blocks found
			MaxCycles [ i ] = 0;
			DoNotCache [ i ] = 0;
			CacheMissCount [ i ] = 0;
		}
	}
	
	// don't forget to invalidate the block
	MaxCycles [ StopBlock ] = 0;
	DoNotCache [ StopBlock ] = 0;
	CacheMissCount [ StopBlock ] = 0;
	*/
}


// regular arithemetic

// *** todo *** no need to save LastModifiedRegister unless instruction is KNOWN to be in a delay slot on run
long Recompiler::ADDU ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rd ].u = r->GPR [ i.Rs ].u + r->GPR [ i.Rt ].u;
	//CHECK_DELAYSLOT ( i.Rd );
	
	int ret = 1;
	
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::ADDU );
			}
			
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].s );
				e->AddRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].s, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	
	return 1;
}

long Recompiler::SUBU ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rd ].u = r->GPR [ i.Rs ].u - r->GPR [ i.Rt ].u;
	//CHECK_DELAYSLOT ( i.Rd );
	
	int ret = 1;
	
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SUBU );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
				e->SubRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	
	return 1;
}

long Recompiler::AND ( Instruction::Format i, u32 Address )
{
	
	int ret = 1;
	
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::AND );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
				e->AndRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RAX );
				
				// note: last modified register can be cached locally in this case, and written back after all instructions encoded
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	
	return 1;
}

long Recompiler::OR ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::OR );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
				e->OrRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	
	return 1;
}

long Recompiler::XOR ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::XOR );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
				e->XorRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NOR ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::NOR );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
				e->OrRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				e->NotReg32 ( RAX );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SLT ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rd ].s = r->GPR [ i.Rs ].s < r->GPR [ i.Rt ].s ? 1 : 0;
	//CHECK_DELAYSLOT ( i.Rd );
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SLT );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->XorRegReg32 ( RCX, RCX );
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
				e->CmpRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				e->Set_L ( RCX );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RCX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SLTU ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rd ].u = r->GPR [ i.Rs ].u < r->GPR [ i.Rt ].u ? 1 : 0;
	//CHECK_DELAYSLOT ( i.Rd );
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SLTU );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->XorRegReg32 ( RCX, RCX );
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
				e->CmpRegMem32 ( RAX, &r->GPR [ i.Rt ].u );
				e->Set_B ( RCX );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RCX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}


////////////////////////////////////////////
// I-Type Instructions (non-interrupt)



long Recompiler::ADDIU ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rt ].s = r->GPR [ i.Rs ].s + i.sImmediate;
	//CHECK_DELAYSLOT ( i.Rt );
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rt )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::ADDIU );
			}
			break;
			
		case 1:
			if ( i.Rt )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].s );
				e->AddRegImm32 ( RAX, i.sImmediate );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rt ].s, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rt );
			}
			Local_LastModifiedReg = i.Rt;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nError encoding ADDIU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::ANDI ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rt ].u = r->GPR [ i.Rs ].u & i.uImmediate;
	//CHECK_DELAYSLOT ( i.Rt );
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rt )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::ANDI );
			}
			break;
			
		case 1:
			if ( i.Rt )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].s );
				e->AndRegImm32 ( RAX, i.uImmediate );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rt ].s, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rt );
			}
			Local_LastModifiedReg = i.Rt;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nError encoding ADDIU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::ORI ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rt ].u = r->GPR [ i.Rs ].u | i.uImmediate;
	//CHECK_DELAYSLOT ( i.Rt );
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rt )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::ORI );
			}
			break;
			
		case 1:
			if ( i.Rt )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].s );
				e->OrRegImm32 ( RAX, i.uImmediate );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rt ].s, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rt );
			}
			Local_LastModifiedReg = i.Rt;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nError encoding ADDIU instruction.\n";
		return -1;
	}
	
	return 1;
}

long Recompiler::XORI ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rt ].u = r->GPR [ i.Rs ].u ^ i.uImmediate;
	//CHECK_DELAYSLOT ( i.Rt );
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rt )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::XORI );
			}
			break;
			
		case 1:
			if ( i.Rt )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].s );
				e->XorRegImm32 ( RAX, i.uImmediate );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rt ].s, RAX );
				
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rt );
			}
			Local_LastModifiedReg = i.Rt;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nError encoding ADDIU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SLTI ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rt ].s = r->GPR [ i.Rs ].s < i.sImmediate ? 1 : 0;
	//CHECK_DELAYSLOT ( i.Rt );
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rt )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SLTI );
			}
			break;
			
		case 1:
			if ( i.Rt )
			{
				e->XorRegReg32 ( RAX, RAX );
				//e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
				e->CmpMemImm32 ( &r->GPR [ i.Rs ].s, i.sImmediate );
				e->Set_L ( RAX );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rt ].u, RAX );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rt );
			}
			Local_LastModifiedReg = i.Rt;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SLTIU ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rt ].u = r->GPR [ i.Rs ].u < ((u32) ((s32) i.sImmediate)) ? 1 : 0;
	//CHECK_DELAYSLOT ( i.Rt );
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rt )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SLTIU );
			}
			break;
			
		case 1:
			if ( i.Rt )
			{
				e->XorRegReg32 ( RAX, RAX );
				//e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
				e->CmpMemImm32 ( &r->GPR [ i.Rs ].s, i.sImmediate );
				e->Set_B ( RAX );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rt ].u, RAX );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rt );
			}
			Local_LastModifiedReg = i.Rt;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::LUI ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rt ].u = ( i.uImmediate << 16 );
	//CHECK_DELAYSLOT ( i.Rt );
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rt )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::LUI );
			}
			break;
			
		case 1:
			if ( i.Rt )
			{
				ret = e->MovMemImm32 ( &r->GPR [ i.Rt ].u, ( i.uImmediate << 16 ) );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rt );
			}
			Local_LastModifiedReg = i.Rt;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LUI instruction.\n";
		return -1;
	}
	return 1;
}







//////////////////////////////////////////////////////////
// Shift instructions



long Recompiler::SLL ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rd ].u = ( r->GPR [ i.Rt ].u << i.Shift );
	//CHECK_DELAYSLOT ( i.Rd );
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SLL );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rt ].u );
				//e->MovRegImm32 ( RCX, (u32) i.Shift );
				e->ShlRegImm32 ( RAX, (u32) i.Shift );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].s, RAX );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SRL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SRL );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rt ].u );
				//e->MovRegImm32 ( RCX, (u32) i.Shift );
				e->ShrRegImm32 ( RAX, (u32) i.Shift );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].s, RAX );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SRA ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SRA );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( 0, &r->GPR [ i.Rt ].u );
				//e->MovRegImm32 ( RCX, (u32) i.Shift );
				e->SarRegImm32 ( 0, (u32) i.Shift );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].s, 0 );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SLLV ( Instruction::Format i, u32 Address )
{
	//r->GPR [ i.Rd ].u = ( r->GPR [ i.Rt ].u << ( r->GPR [ i.Rs ].u & 0x1f ) );
	//CHECK_DELAYSLOT ( i.Rd );
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SLLV );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rt ].u );
				e->MovRegFromMem32 ( RCX, &r->GPR [ i.Rs ].u );
				e->ShlRegReg32 ( RAX );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].s, RAX );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SRLV ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SRLV );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rt ].u );
				e->MovRegFromMem32 ( RCX, &r->GPR [ i.Rs ].u );
				e->ShrRegReg32 ( RAX );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].s, RAX );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SRAV ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			if ( i.Rd )
			{
				// load arguments
				e->LoadImm32 ( RCX, i.Value );
				ret = e->Call ( Instruction::Execute::SRAV );
			}
			break;
			
		case 1:
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rt ].u );
				e->MovRegFromMem32 ( RCX, &r->GPR [ i.Rs ].u );
				e->SarRegReg32 ( RAX );
				e->MovRegToMem32 ( &r->GPR [ i.Rd ].s, RAX );
				//ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	} // end switch ( OpLevel )
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDU instruction.\n";
		return -1;
	}
	return 1;
}



////////////////////////////////////////////
// Jump/Branch Instructions



long Recompiler::J ( Instruction::Format i, u32 Address )
{
	//r->DelaySlot0.Instruction = i;
	//r->DelaySlot0.cb = r->_cb_Jump;
	//r->Status.DelaySlot_Valid |= 0x1;
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			// load arguments
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::J );
			break;
			
		case 1:
			return -1;
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding J instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::JR ( Instruction::Format i, u32 Address )
{
	//r->DelaySlot0.Instruction = i;
	//r->DelaySlot0.cb = r->_cb_JumpRegister;
	//r->DelaySlot0.Data = r->GPR [ i.Rs ].u & ~3;
	//r->Status.DelaySlot_Valid |= 0x1;
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::JR );
			break;
			
		case 1:
			return -1;
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding JR instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::JAL ( Instruction::Format i, u32 Address )
{
	//r->DelaySlot0.Instruction = i;
	//r->DelaySlot0.cb = r->_cb_Jump;
	//r->Status.DelaySlot_Valid |= 0x1;
	//r->GPR [ 31 ].u = r->PC + 8;
	//CHECK_DELAYSLOT ( 31 );
	int ret = 1;
	
	// for now, stop encoding before this instruction
	bStopEncodingBefore = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::JAL );
			break;
			
		case 1:
			return -1;
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			e->MovMemImm32 ( &r->GPR [ 31 ].u, Address + 8 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 31 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding JAL instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::JALR ( Instruction::Format i, u32 Address )
{
	//r->DelaySlot0.Instruction = i;
	//r->DelaySlot0.Data = r->GPR [ i.Rs ].u & ~3;
	//r->DelaySlot0.cb = r->_cb_JumpRegister;
	//r->Status.DelaySlot_Valid |= 0x1;
	//r->GPR [ i.Rd ].u = r->PC + 8;
	//CHECK_DELAYSLOT ( i.Rd );
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingBefore = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::JALR );
			
			// make sure Rd is not r0
			if ( !i.Rd )
			{
				// make sure r0 stays at zero
				e->MovMemImm32 ( &r->GPR [ 0 ].u, 0 );
			}
			break;
			
		case 1:
			return -1;
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			if ( i.Rd )
			{
				e->MovMemImm32 ( &r->GPR [ i.Rd ].u, Address + 8 );
				ret = e->MovMemImm32 ( &r->LastModifiedRegister, i.Rd );
			}
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding JALR instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BEQ ( Instruction::Format i, u32 Address )
{
	//if ( r->GPR [ i.Rs ].u == r->GPR [ i.Rt ].u )
	//{
	//	r->DelaySlot0.Instruction = i;
	//	r->DelaySlot0.cb = r->_cb_Branch;
	//	r->Status.DelaySlot_Valid |= 0x1;
	//}
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BEQ );
			break;
			
		case 1:
			return -1;
			e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->CmpRegMem32 ( 0, &r->GPR [ i.Rt ].u );
			e->Jmp8_NE ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BEQ instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BNE ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BNE );
			break;
			
		case 1:
			return -1;
			e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->CmpRegMem32 ( 0, &r->GPR [ i.Rt ].u );
			e->Jmp8_E ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BNE instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BLEZ ( Instruction::Format i, u32 Address )
{
	//if ( r->GPR [ i.Rs ].s <= 0 )
	//{
	//	// next instruction is in the branch delay slot
	//	r->DelaySlot0.Instruction = i;
	//	r->DelaySlot0.cb = r->_cb_Branch;
	//	r->Status.DelaySlot_Valid |= 0x1;
	//}
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BLEZ );
			break;
			
		case 1:
			return -1;
			//e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->CmpMemImm32 ( &r->GPR [ i.Rs ].u, 0 );
			e->Jmp8_G ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BLEZ instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BGTZ ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BGTZ );
			break;
			
		case 1:
			return -1;
			//e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->CmpMemImm32 ( &r->GPR [ i.Rs ].u, 0 );
			e->Jmp8_LE ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BGTZ instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BLTZ ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BLTZ );
			break;
			
		case 1:
			return -1;
			//e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->CmpMemImm32 ( &r->GPR [ i.Rs ].u, 0 );
			e->Jmp8_GE ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BLTZ instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BGEZ ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BGEZ );
			break;
			
		case 1:
			return -1;
			//e->MovRegFromMem32 ( 0, &r->GPR [ i.Rs ].u );
			e->CmpMemImm32 ( &r->GPR [ i.Rs ].u, 0 );
			e->Jmp8_L ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			ret = e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BGEZ instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BLTZAL ( Instruction::Format i, u32 Address )
{
	//if ( r->GPR [ i.Rs ].s < 0 )
	//{
	//	r->DelaySlot0.Instruction = i;
	//	r->DelaySlot0.cb = r->_cb_Branch;
	//	r->Status.DelaySlot_Valid |= 0x1;
	//}
	//r->GPR [ 31 ].u = r->PC + 8;
	//CHECK_DELAYSLOT ( 31 );
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingBefore = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BLTZAL );
			break;
			
		case 1:
			return -1;
			e->CmpMemImm32 ( &r->GPR [ i.Rs ].u, 0 );
			e->Jmp8_GE ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			e->MovMemImm32 ( &r->GPR [ 31 ].u, Address + 8 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 31 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BLTZAL instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BGEZAL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingBefore = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BGEZAL );
			break;
			
		case 1:
			return -1;
			e->CmpMemImm32 ( &r->GPR [ i.Rs ].u, 0 );
			e->Jmp8_L ( 0, 0 );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			e->SetJmpTarget8 ( 0 );
			e->MovMemImm32 ( &r->GPR [ 31 ].u, Address + 8 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 31 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BGEZAL instruction.\n";
		return -1;
	}
	return 1;
}


/////////////////////////////////////////////////////////////
// Multiply/Divide Instructions

// *** todo *** don't forget to fix CPU wait function to include storing from store buffer
long Recompiler::MULT ( Instruction::Format i, u32 Address )
{
	//if ( r->MulDiv_BusyUntil_Cycle > r->CycleCount )
	//{
	//	r->CycleCount = r->MulDiv_BusyUntil_Cycle;
	//}
	//r->MulDiv_BusyUntil_Cycle = r->CycleCount + c_iMultiplyCycles_Slow;
	//if ( r->GPR [ i.Rs ].s < 0x800 && r->GPR [ i.Rs ].s >= -0x800 )
	//{
	//	r->MulDiv_BusyUntil_Cycle = r->CycleCount + c_iMultiplyCycles_Fast;
	//}
	//else if ( r->GPR [ i.Rs ].s < 0x100000 && r->GPR [ i.Rs ].s >= -0x100000 )
	//{
	//	r->MulDiv_BusyUntil_Cycle = r->CycleCount + c_iMultiplyCycles_Med;
	//}
	// multiply signed Lo,Hi = rs * rt
	//r->HiLo.sValue = ((s64) (r->GPR [ i.Rs ].s)) * ((s64) (r->GPR [ i.Rt ].s));
	
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MULT );
			break;
			
		case 1:
			return -1;
			//e->MovRegFromMem64 ( RAX, &r->CycleCount );
			//e->CmpRegMem64 ( RAX, &r->MulDiv_BusyUntil_Cycle );
			//e->CmovBRegMem64 ( RAX, &r->MulDiv_BusyUntil_Cycle );
			//e->MovRegFromMem32 ( RCX, &r->GPR [ i.Rs ].s )
			//e->NotReg32 ( RCX );
			//e->MovRegImm32 ( RDX, 6 );
			//e->CmpRegImm32 ( RCX, 0x800 );
			//e->CmovBReg
			
			//ret = e->MovRegToMem64 ( &r->CycleCount, RAX );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MULT instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::MULTU ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MULTU );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MULTU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::DIV ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::DIV );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding DIV instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::DIVU ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::DIVU );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding DIVU instruction.\n";
		return -1;
	}
	return 1;
}



long Recompiler::MFHI ( Instruction::Format i, u32 Address )
{
	//if ( r->MulDiv_BusyUntil_Cycle > r->CycleCount )
	//{
	//	r->CycleCount = r->MulDiv_BusyUntil_Cycle;
	//}
	//r->GPR [ i.Rd ].u = r->HiLo.uHi;
	//CHECK_DELAYSLOT ( i.Rd );
	
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MFHI );
			// make sure Rd is not r0
			if ( !i.Rd )
			{
				// make sure r0 stays at zero
				ret = e->MovMemImm32 ( &r->GPR [ 0 ].u, 0 );
			}
			break;
			
		case 1:
			e->MovRegFromMem64 ( RAX, &r->CycleCount );
			e->CmpRegMem64 ( RAX, &r->MulDiv_BusyUntil_Cycle );
			e->CmovBRegMem64 ( RAX, &r->MulDiv_BusyUntil_Cycle );
			ret = e->MovRegToMem64 ( &r->CycleCount, RAX );
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->HiLo.uHi );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RAX );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MFHI instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::MFLO ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MFLO );
			// make sure Rd is not r0
			if ( !i.Rd )
			{
				// make sure r0 stays at zero
				ret = e->MovMemImm32 ( &r->GPR [ 0 ].u, 0 );
			}
			break;
			
		case 1:
			e->MovRegFromMem64 ( RAX, &r->CycleCount );
			e->CmpRegMem64 ( RAX, &r->MulDiv_BusyUntil_Cycle );
			e->CmovBRegMem64 ( RAX, &r->MulDiv_BusyUntil_Cycle );
			ret = e->MovRegToMem64 ( &r->CycleCount, RAX );
			if ( i.Rd )
			{
				e->MovRegFromMem32 ( RAX, &r->HiLo.uLo );
				ret = e->MovRegToMem32 ( &r->GPR [ i.Rd ].u, RAX );
			}
			Local_LastModifiedReg = i.Rd;
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MFLO instruction.\n";
		return -1;
	}
	return 1;
}




long Recompiler::MTHI ( Instruction::Format i, u32 Address )
{
	//r->HiLo.uHi = r->GPR [ i.Rs ].u;
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MTHI );
			break;
			
		case 1:
			e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
			ret = e->MovRegToMem32 ( &r->HiLo.uHi, RAX );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MTHI instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::MTLO ( Instruction::Format i, u32 Address )
{
	//r->HiLo.uLo = r->GPR [ i.Rs ].u;
	
	int ret = 1;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MTLO );
			break;
			
		case 1:
			e->MovRegFromMem32 ( RAX, &r->GPR [ i.Rs ].u );
			ret = e->MovRegToMem32 ( &r->HiLo.uLo, RAX );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MTLO instruction.\n";
		return -1;
	}
	return 1;
}




long Recompiler::RFE ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::RFE );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding RFE instruction.\n";
		return -1;
	}
	return 1;
}




////////////////////////////////////////////////////////
// Instructions that can cause Synchronous Interrupts //
////////////////////////////////////////////////////////


long Recompiler::ADD ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding before this instruction
	// so if this is not the only instruction in run, then don't encode it
	bStopEncodingBefore = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::ADD );
			if ( !i.Rd )
			{
				// make sure r0 stays at zero
				ret = e->MovMemImm32 ( &r->GPR [ 0 ].u, 0 );
			}
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADD instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::ADDI ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding before this instruction
	// so if this is not the only instruction in run, then don't encode it
	bStopEncodingBefore = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::ADDI );
			if ( !i.Rt )
			{
				// make sure r0 stays at zero
				ret = e->MovMemImm32 ( &r->GPR [ 0 ].u, 0 );
			}
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding ADDI instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SUB ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding before this instruction
	// so if this is not the only instruction in run, then don't encode it
	bStopEncodingBefore = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SUB );
			if ( !i.Rd )
			{
				// make sure r0 stays at zero
				ret = e->MovMemImm32 ( &r->GPR [ 0 ].u, 0 );
			}
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SUB instruction.\n";
		return -1;
	}
	return 1;
}




long Recompiler::SYSCALL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SYSCALL );
			Local_NextPCModified = true;
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SYSCALL instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::BREAK ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::BREAK );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding BREAK instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::Invalid ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::Invalid );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding Invalid instruction.\n";
		return -1;
	}
	return 1;
}





long Recompiler::MFC0 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MFC0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MFC0 instruction.\n";
		return -1;
	}
	return 1;
}




long Recompiler::MTC0 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MTC0 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MTC0 instruction.\n";
		return -1;
	}
	return 1;
}




long Recompiler::MFC2 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MFC2 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MFC2 instruction.\n";
		return -1;
	}
	return 1;
}




long Recompiler::MTC2 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MTC2 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MTC2 instruction.\n";
		return -1;
	}
	return 1;
}




long Recompiler::CFC2 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::CFC2 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding CFC2 instruction.\n";
		return -1;
	}
	return 1;
}



long Recompiler::CTC2 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::CTC2 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding CTC2 instruction.\n";
		return -1;
	}
	return 1;
}




// Load/Store - will need to use address translation to get physical addresses when needed

//////////////////////////////////////////////////////////////////////////
// store instructions

// store instructions
long Recompiler::SB ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SB );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SB instruction.\n";
		return -1;
	}
	return 1;
}





long Recompiler::SH ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SH );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SH instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SW ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SW );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SW instruction.\n";
		return -1;
	}
	return 1;
}


long Recompiler::SWC2 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SWC2 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SWC2 instruction.\n";
		return -1;
	}
	return 1;
}



long Recompiler::SWL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SWL );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SWL instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SWR ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SWR );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SWR instruction.\n";
		return -1;
	}
	return 1;
}



/////////////////////////////////////////////////
// load instructions

// load instructions with delay slot
// *** todo *** it is also possible to this and just process load after load delay slot has executed - would still need previous load address before delay slot
// *** todo *** could also skip delay slot zero and put straight into delay slot 1 after next instruction, or just process load delay slot after next instruction
long Recompiler::LB ( Instruction::Format i, u32 Address )
{
	//LoadAddress = r->GPR [ i.Base ].s + i.sOffset;
	//r->DelaySlot0.Instruction = i;
	//r->DelaySlot0.Data = LoadAddress;
	//r->DelaySlot0.cb = LB_DelaySlot_Callback_Bus;
	//r->Status.DelaySlot_Valid |= 0x1;
	//r->LastModifiedRegister = 255;
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LB );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LB instruction.\n";
		return -1;
	}
	return 1;
}






long Recompiler::LH ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LH );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LB instruction.\n";
		return -1;
	}
	return 1;
}








long Recompiler::LW ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LW );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LB instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::LBU ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LBU );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LBU instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::LHU ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LHU );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LB instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::LWC2 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LWC2 );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LWC2 instruction.\n";
		return -1;
	}
	return 1;
}




// load instructions without load-delay slot
long Recompiler::LWL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LWL );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LB instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::LWR ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::LWR );
			break;
			
		case 1:
			return -1;
			e->MovRegImm32 ( 0, i.sOffset );
			e->AddRegMem32 ( 0, &r->GPR [ i.Base ].u );
			e->MovMemImm32 ( &r->DelaySlot0.Instruction.Value, i.Value );
			e->MovRegToMem32 ( &r->DelaySlot0.Data, 0 );
			e->OrMemImm64 ( &r->Status.Value, 1 );
			ret = e->MovMemImm32 ( &r->LastModifiedRegister, 255 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding LWR instruction.\n";
		return -1;
	}
	return 1;
}




















///////////////////////////
// GTE instructions

long Recompiler::COP2 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::COP2 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding COP2 instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::RTPS ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::RTPS );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding RTPS instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NCLIP ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::NCLIP );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding NCLIP instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::OP ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::OP );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding OP instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::DPCS ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::DPCS );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding DPCS instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::INTPL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::INTPL );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding INTPL instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::MVMVA ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::MVMVA );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding MVMVA instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NCDS ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::NCDS );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding NCDS instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::CDP ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::CDP );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding CDP instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NCDT ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::NCDT );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding NCDT instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NCCS ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::NCCS );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding NCCS instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::CC ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::CC );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding CC instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NCS ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::NCS );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding NCS instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NCT ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::NCT );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding NCT instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::SQR ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::SQR );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding SQR instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::DCPL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::DCPL );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding DCPL instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::DPCT ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::DPCT );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding DPCT instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::AVSZ3 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::AVSZ3 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding AVSZ3 instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::AVSZ4 ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::AVSZ4 );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding AVSZ4 instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::RTPT ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::RTPT );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding RTPT instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::GPF ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::GPF );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding GPF instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::GPL ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::GPL );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding GPL instruction.\n";
		return -1;
	}
	return 1;
}

long Recompiler::NCCT ( Instruction::Format i, u32 Address )
{
	int ret = 1;
	// for now, stop encoding after this instruction
	bStopEncodingAfter = true;
	switch ( OpLevel )
	{
		case 0:
			e->LoadImm32 ( RCX, i.Value );
			ret = e->Call ( Instruction::Execute::NCCT );
			break;
			
		default:
			return -1;
			break;
	}
	
	if ( !ret )
	{
		cout << "\nx64 Recompiler: Error encoding NCCT instruction.\n";
		return -1;
	}
	return 1;
}



static const Recompiler::Function Recompiler::FunctionList []
{
	// instructions on both R3000A and R5900
	Recompiler::Invalid,
	Recompiler::J, Recompiler::JAL, Recompiler::JR, Recompiler::JALR, Recompiler::BEQ, Recompiler::BNE, Recompiler::BGTZ, Recompiler::BGEZ,
	Recompiler::BLTZ, Recompiler::BLEZ, Recompiler::BGEZAL, Recompiler::BLTZAL, Recompiler::ADD, Recompiler::ADDI, Recompiler::ADDU, Recompiler::ADDIU,
	Recompiler::SUB, Recompiler::SUBU, Recompiler::MULT, Recompiler::MULTU, Recompiler::DIV, Recompiler::DIVU, Recompiler::AND, Recompiler::ANDI,
	Recompiler::OR, Recompiler::ORI, Recompiler::XOR, Recompiler::XORI, Recompiler::NOR, Recompiler::LUI, Recompiler::SLL, Recompiler::SRL,
	Recompiler::SRA, Recompiler::SLLV, Recompiler::SRLV, Recompiler::SRAV, Recompiler::SLT, Recompiler::SLTI, Recompiler::SLTU, Recompiler::SLTIU,
	Recompiler::LB, Recompiler::LBU, Recompiler::LH, Recompiler::LHU, Recompiler::LW, Recompiler::LWL, Recompiler::LWR, Recompiler::SB,
	Recompiler::SH, Recompiler::SW, Recompiler::SWL, Recompiler::SWR, Recompiler::MFHI, Recompiler::MTHI, Recompiler::MFLO, Recompiler::MTLO,
	Recompiler::MFC0, Recompiler::MTC0, Recompiler::CFC2, Recompiler::CTC2, Recompiler::SYSCALL, Recompiler::BREAK,
	
	// instructions on R3000A ONLY
	Recompiler::MFC2, Recompiler::MTC2, Recompiler::LWC2, Recompiler::SWC2, Recompiler::RFE,
	Recompiler::RTPS, Recompiler::RTPT, Recompiler::CC, Recompiler::CDP, Recompiler::DCPL, Recompiler::DPCS, Recompiler::DPCT, Recompiler::NCS,
	Recompiler::NCT, Recompiler::NCDS, Recompiler::NCDT, Recompiler::NCCS, Recompiler::NCCT, Recompiler::GPF, Recompiler::GPL, Recompiler::AVSZ3,
	Recompiler::AVSZ4, Recompiler::SQR, Recompiler::OP, Recompiler::NCLIP, Recompiler::INTPL, Recompiler::MVMVA
};
